"""Implementaions for downloading and resampling ERA5 data."""

import argparse
import glob
import os
import shutil
import time
from os.path import basename, isfile, join
from typing import Any, Dict, List, Optional

import dask
import numpy as np
import pandas as pd
import xarray as xr
from dask.diagnostics import ProgressBar

_JOULE_PER_M_SQ = "J m**-2"

# define global vars
dt_era5t = pd.to_timedelta(95, unit="d")
era5_variables = {
    "ssrd": "surface_solar_radiation_downwards",
    "t2m": "2m_temperature",
    "tisr": "toa_incident_solar_radiation",
    "tp": "total_precipitation",
    "msl": "mean_sea_level_pressure",
    "cape": "convective_available_potential_energy",
    "tcwv": "total_column_water_vapour",
    "shww": "significant_height_of_wind_waves",
    "ro": "runoff",
    "pev": "potential_evaporation",
    "u10": "10m_u_component_of_wind",
    "v10": "10m_v_component_of_wind",
    "d2m": "2m_dewpoint_temperature",
    "ssr": "surface_net_solar_radiation",
    "tcc": "total_cloud_cover",
}
daily_attrs = {
    "tp": {
        "units": "mm d**-1",  # NOTE unit conversion m to mm
        "long_name": "Total precipitation",
    },
    "tmin": {"units": "K", "long_name": "2 meter mininum temperature"},
    "tmax": {"units": "K", "long_name": "2 meter maximum temperature"},
    "t2m": {"units": "K", "long_name": "2 meter mean temperature"},
    "d2m": {"units": "K", "long_name": "2 meter dewpoint temperature"},
    "msl": {
        "units": "Pa",
        "long_name": "Mean sea level pressure",
        "standard_name": "air_pressure_at_mean_sea_level",
    },
    "tisr": {"units": _JOULE_PER_M_SQ, "long_name": "TOA incident solar radiation"},
    "ssrd": {
        "units": _JOULE_PER_M_SQ,
        "long_name": "Surface solar radiation downwards",
        "standard_name": "surface_downwelling_shortwave_flux_in_air",
    },
    "tcwv": {
        "units": "kg m**-2",
        "long_name": "Total column water vapour",
    },
    "u10": {
        "units": "m s**-1",
        "long_name": "Neutral wind at 10 m u-component",
        "standard_name": "10m_u_component_of_wind",
    },
    "v10": {
        "units": "m s**-1",
        "long_name": "Neutral wind at 10 m v-component",
        "standard_name": "10m_v_component_of_wind",
    },
    "ssr": {
        "units": _JOULE_PER_M_SQ,
        "long_name": "Surface net solar radiation",
        "standard_name": "surface_net_downward_shortwave_flux",
    },
    "tcc": {
        "units": "-",
        "long_name": "Total cloud cover",
        "standard_name": "cloud_area_fraction",
    },
}


def download_era5_year(
    write_path: str,
    variable: str,
    year: int,
    months: Optional[List[str]] = None,
    days: Optional[List[str]] = None,
) -> None:
    """Download a ERA5 variable for a single year from Copernicus Climate Data Store.

    Arguments
    ---------
    write_path: (str)
        The output path to save the downloaded file.
    variable: (str)
        The CDS variable name to download.
    year: (int)
        The year to download the data for.
    months: (list)
        List of months to download data for. Use None to download data for all months.
    days: (list)
        List of days to download data for. Use None to downloads data for all days.
    """
    import cdsapi

    if not isfile(write_path):
        # Set default months, days and hours options
        _months = [f"{m:02d}" for m in range(1, 13)]
        _days = [f"{d:02d}" for d in range(1, 32)]
        _hrs = [f"{h:02d}:00" for h in range(24)]

        if months is None:
            months = _months
        if days is None:
            days = _days

        dataset = "reanalysis-era5-single-levels"
        if int(year) < 1979:
            dataset = f"{dataset}-complete-preliminary-back-extension"

        # Download ERA5 data
        c = cdsapi.Client()
        c.retrieve(
            dataset,
            {
                "product_type": "reanalysis",
                "format": "netcdf",
                "variable": variable,
                "year": [year],
                "month": months,
                "day": days,
                "time": _hrs,
            },
            write_path,
        )


def flatten_era5_temp(
    path,
    nodata=-9999,
    chunks=None,
    dask_kwargs=None,
) -> None:
    """Flatten ERA5 output to remove expver dimension with ERA5T data."""
    chunks = chunks or {"time": 30, "latitude": 250, "longitude": 480}
    dask_kwargs = dask_kwargs or {}
    with xr.open_dataset(path, chunks=chunks) as ds:
        era5t = "expver" in ds.coords

    if era5t:
        print(f"flattening {basename(path)}")
        write_path = path.replace(".nc", "_flat.nc")
        with xr.open_dataset(path, chunks=chunks) as ds:
            # expver5 (ERA5T) contains NaNs when out of range,
            # while expver1 might contain large odd values
            # https://confluence.ecmwf.int/display/CUSF/ERA5+CDS+requests+which+return+a+mixture+of+ERA5+and+ERA5T+data
            ds_out = ds.sel(expver=5).combine_first(ds.sel(expver=1)).fillna(nodata)
            chunksizes = tuple([s[0] for s in ds_out.chunks.values()])
            e0 = {
                "zlib": True,
                "dtype": "float32",
                "_FillValue": nodata,
                "chunksizes": chunksizes,
            }
            encoding = {var: e0 for var in ds_out.data_vars}
            obj = ds_out.to_netcdf(
                write_path, encoding=encoding, mode="w", compute=False
            )
            with ProgressBar():
                obj.compute(**dask_kwargs)
        if isfile(write_path):
            # replace original file
            os.unlink(path)
            shutil.move(write_path, path)


# resample variable for single year!
def resample_year(
    year: int,
    ddir: str,
    outdir: str,
    var: str,
    decimals: Optional[int] = None,
    nodata: int = -9999,
    chunks: Optional[Dict[Any, Any]] = None,
    dask_kwargs: Optional[Dict[str, Any]] = None,
) -> List[str]:
    """Resample hourly variables to daily timestep.

    The data is saved with the time labels at the end of the timestep.
    By default the data is aggregated using the mean.

    Exceptions:
    - the unit of tp is changed from m/hr to mm/day
    - from t2m, the mean, min and max daily values are saved

    Parameters
    ----------
    year : int
        year
    ddir : str
        Root of hourly nc files with path format
        {ddir}/{var}/era5_{var}_{year}_hourly.nc
    outdir : str
        Temporary output folder
    var : str
        input variable
    decimals : int, optional
        _description_, by default None
    nodata : int, optional
        _description_, by default -9999
    chunks : _type_, optional
        _description_, by default {'time':30, 'latitude':250, 'longitude':480}
    dask_kwargs : dict, optional
        _description_, by default {}
    """
    chunks = chunks or {"time": 30, "latitude": 250, "longitude": 480}
    dask_kwargs = dask_kwargs or {}

    # nc out settings
    chunksizes = tuple([s for s in chunks.values()])
    e0 = {
        "zlib": True,
        "dtype": "float32",
        "_FillValue": nodata,
        "chunksizes": chunksizes,
    }

    paths = []
    # read hourly data for year and year-1!
    for _year in [year - 1, year]:
        path = join(ddir, var, f"era5_{var:s}_{year:d}_hourly.nc")
        if isfile(path):
            paths.append(path)
    kwargs = dict(
        compat="no_conflicts",
        parallel=True,
        decode_times=True,
    )
    ds = xr.open_mfdataset(paths, chunks=chunks, **kwargs)
    assert "expver" not in ds.coords
    ds = ds.sel(time=slice(f"{year - 1}-12-31 01:00", f"{year}-12-31 00:00"))
    if ds["time"][-1].dt.hour != 0:  # clip incomplete days
        month, day = ds["time"][-1].dt.month.item(), ds["time"][-1].dt.day.item()
        ds = ds.sel(
            time=slice(f"{year - 1}-12-31 01:00", f"{year}-{month}-{day} 00:00")
        )

    # resample to daily freq
    kwargs = dict(time="1D", label="right", closed="right")
    dvars = {}
    if "tp" in ds:
        # NOTE unit conversion m to mm
        dvars["tp"] = ds["tp"].resample(**kwargs).sum("time") * 1000
        dvars["tp"].attrs.update(daily_attrs.get("tp", {}))
    if "t2m" in ds:
        dvars["tmin"] = (
            ds[["t2m"]].resample(**kwargs).min("time").rename({"t2m": "tmin"})
        )
        dvars["tmin"].attrs.update(daily_attrs.get("tmin", {}))
        dvars["tmax"] = (
            ds[["t2m"]].resample(**kwargs).max("time").rename({"t2m": "tmax"})
        )
        dvars["tmax"].attrs.update(daily_attrs.get("tmax", {}))
    for var in ds.data_vars.keys():
        if var in dvars:
            continue
        dvars[var] = ds[var].resample(**kwargs).mean("time")
        dvars[var].attrs.update(daily_attrs.get(var, {}))
    ds_out = xr.merge(dvars.values()).chunk(chunks).fillna(nodata)
    if decimals:
        ds_out = ds_out.round(decimals=decimals)

    # write
    paths = []
    for var in ds_out.data_vars.keys():
        print(f"{var}: {year} - {ds_out.time.size} days")
        write_path = join(outdir, f"era5_{var:s}_{year:d}_daily.nc")
        paths.append(write_path)
        if isfile(write_path):
            continue
        obj = ds_out[[var]].to_netcdf(
            write_path, encoding={var: e0}, mode="w", compute=False
        )
        with ProgressBar():
            obj.compute(**dask_kwargs)

    return paths


def move_replace(src: str, dst: str, timeout: int = 300) -> None:
    """Try replacing old file which might be locked."""
    if not isfile(src):
        return
    if not os.path.isdir(os.path.dirname(dst)):
        os.makedirs(os.path.dirname(dst))
    while isfile(dst):
        try:
            os.unlink(dst)
        except PermissionError:
            print(f"FAILED deleting {dst} (retry in {timeout} sec)")
            time.sleep(timeout)
    os.rename(src, dst)


def get_last_timestep_nc(paths: list) -> pd.Timedelta:
    """Get the last timestep from a list of NetCDF files.

    Parameters
    ----------
    paths : list
        List of file paths or a glob pattern to match the files.

    Returns
    -------
    pd.Timedelta
        The time duration from the reference time to the last timestep.

    Raises
    ------
    ValueError
        If no files are found matching the provided file paths or glob pattern.
    """
    if len(glob.glob(paths)) == 0:
        raise ValueError(f"Files not found: {paths}")
    with xr.open_mfdataset(paths, chunks="auto") as ds:
        t0 = pd.to_datetime(ds["time"][-1].values)
    return t0


def append_zarr(
    store: str,
    ds: xr.Dataset,
    chunks_out: dict,
    append_dim: str = "time",
) -> None:
    """Write/append/overwrite data to a zarr store.

    Parameters
    ----------
    store : str
        path to zarr store
    ds : xarray.Dataset
        Dataset to append
    chunks_out : dict
        zarr chunks (only used for new store)
    append_dim : str, optional
        dimension along wicht to append, by default 'time'
    """
    import zarr
    from cftime import date2num
    from numcodecs import Blosc

    compressor = Blosc(cname="zstd", clevel=3, shuffle=1)

    encoding = {}
    ds.encoding = {}
    for v in ds.data_vars:
        ds[v].encoding = {}
        if compressor is not None:
            encoding[v] = {"compressor": compressor}

    # pre-allocate new zarr store
    if not os.path.isdir(store):
        ds.chunk(chunks_out).to_zarr(
            store,
            compute=False,
            consolidated=False,
            encoding=encoding,
        )

    ds0 = xr.open_zarr(store, consolidated=False)
    for v in ds.data_vars:
        # pre-allocate new zarr variables
        if v not in ds0:
            _da0 = ds0[list(ds0.data_vars.keys())[0]]
            _ds = xr.DataArray(
                name=v,
                data=dask.array.empty_like(_da0),
                coords=_da0.coords,
                dims=_da0.dims,
                attrs=ds[v].attrs,
            ).to_dataset()
            _ds.chunk(chunks_out).to_zarr(
                store,
                mode="a",
                compute=False,
                consolidated=False,
                encoding={v: encoding.get(v, {})},
            )

    # make sure ds does not start before ds0
    ds = ds.sel(
        {append_dim: slice(ds0[append_dim].values[0], ds[append_dim].values[-1])}
    )
    # get start- and end- time of both datsets and calculate offset
    zgroup = zarr.open_group(store)
    t0 = zgroup[append_dim][0]
    dt = int(np.diff(zgroup[append_dim][:2])[0])
    time_encoding = ds0[append_dim].encoding
    _t0 = pd.Timestamp(ds[append_dim].values[0]).to_pydatetime()
    _t1 = pd.Timestamp(ds[append_dim].values[-1]).to_pydatetime()
    _t0, _t1 = date2num([_t0, _t1], time_encoding["units"], time_encoding["calendar"])
    offset = int((_t0 - t0) / dt)

    # check if time dimension needs to be extended
    if _t1 > zgroup[append_dim][-1]:
        dates = np.arange(t0, _t1 + 1, dt, dtype=zgroup[append_dim].dtype)
        zgroup[append_dim].resize(dates.size)
        zgroup[append_dim][:] = dates
        # expand time on all variables
        for v in ds0.data_vars:
            if append_dim in ds0[v].dims:
                assert ds0[v].get_axis_num(append_dim) == 0
                arr = zgroup[v]
                shape = list(arr.shape)
                shape[0] = dates.size
                zgroup[v].resize(shape)

    # align dims and coords
    # re-read ds as new variables might have been appended
    ds0 = xr.open_zarr(store, consolidated=False)
    for v in ds.data_vars:
        ds[v] = ds[v].transpose(*ds0[v].dims)
        for _idim, dim in enumerate(ds.dims):
            if dim == append_dim:
                continue
            assert ds[v][dim].shape == ds0[v][dim].shape
            if not np.allclose(ds0[v][dim], ds[v][dim]):
                raise ValueError("ds0 and ds don't have the same dimensions")

    # write output using to_zarr with region argument
    # drop dims (already written)
    region_out = {append_dim: slice(offset, offset + ds[append_dim].size)}
    ds.drop(ds0[v].dims).to_zarr(
        store,
        region=region_out,
        compute=True,
        safe_chunks=True,
        consolidated=False,
    )
    # consolidate metadata
    zarr.consolidate_metadata(store)


def update_hourly_nc(
    outdir: str,
    ddir: str,
    variables: list,
    start_year: int = None,
    end_year: int = None,
    dt_era5t: pd.Timedelta = dt_era5t,
    move_to_ddir: bool = False,
    dask_kwargs: dict = None,
) -> None:
    """Update the hourly by downloading the latest data from the CDS.

    Parameters
    ----------
    outdir : str
        temporary output folder
    ddir : str
        Root of hourly nc files with path format
        {ddir}/{var}/era5_{var}_{year}_hourly.nc
    variables : list
        list of variable names to update
    start_year : int, optional
        start year, by default None and read from last timestep of hourly nc files
        minus dt_era5t
    end_year : int, optional
        end year, by default None and based on todays' date
    dt_era5t : pd.Timedelta, optional
        time period of ERA5T data which is overwritten by new data, by default dt_era5t
    move_to_ddir : bool, optional
        Move files to root of hourly nc, by default False
    dask_kwargs
            Additional key-word arguments are passed to dask
    """
    dask_kwargs = dask_kwargs or {}
    t1 = pd.to_datetime("today").to_numpy()
    if end_year:
        t1 = pd.to_datetime(f"{end_year}0101")
    if start_year:
        t0 = pd.to_datetime(f"{start_year}0101")

    # download
    jobs = []
    for var in variables:
        if start_year is None:  # get last date of hourly file
            paths = join(ddir, var, f"era5_{var}_*_hourly.nc")
            t0 = get_last_timestep_nc(paths)
            if (t0 > t1) or ((t1 - t0) < pd.to_timedelta(7, unit="d")):
                continue  ## skip if last date within 7 days
            t0 = t0 - dt_era5t
        years = np.unique(pd.date_range(t0, t1, freq="1d").year).tolist()

        # download complete years
        for year in years:
            write_path = join(outdir, f"era5_{var}_{year}_hourly.nc")
            job = dask.delayed(download_era5_year)(
                write_path, era5_variables[var], year=year
            )
            jobs.append(job)
    dask.compute(*jobs, **dask_kwargs)

    # flatten if the data contains both ERA5 & ERA5T data
    for var in variables:
        paths = glob.glob(join(outdir, f"era5_{var}_*_hourly.nc"))
        for p in paths:
            flatten_era5_temp(p, dask_kwargs=dask_kwargs)

    # write files from tmpdir to ddir
    if move_to_ddir:
        for var in variables:
            paths = glob.glob(join(outdir, f"era5_{var}_*_hourly.nc"))
            for p in paths:
                move_replace(p, join(ddir, var, basename(p)))


def update_daily_nc(
    outdir: str,
    ddir_hour: str,
    ddir_day: str,
    variables: list,
    start_year: int = None,
    end_year: int = None,
    dt_era5t: pd.Timedelta = dt_era5t,
    move_to_ddir: bool = False,
    dask_kwargs: dict = None,
) -> None:
    """Update the daily nc files based on hourly files.

    Parameters
    ----------
    outdir : str
        temporary output folder
    ddir_hour : str
        Root of hourly nc files with path format
        {ddir}/{var}/era5_{var}_{year}_hourly.nc
    ddir_day : str
        Root of daily nc files with path format {ddir}/{var}/era5_{var}_{year}_daily.nc
    variables : list
        list of variable names to update
    start_year : int, optional
        start year, by default None and read from last timestep of daily nc
        files minus dt_era5t
    end_year : int, optional
        end year, by default None and read from last timestep of hourly nc files
    dt_era5t : pd.Timedelta, optional
        time period of ERA5T data which is overwritten by new data, by default dt_era5t
    move_to_ddir : bool, optional
        Move files to root of daily nc, by default False
    dask_kwargs
            Additional key-word arguments are passed to dask
    """
    dask_kwargs = dask_kwargs or {}
    if end_year:
        t1 = pd.to_datetime(f"{end_year}0101")
    if start_year:
        t0 = pd.to_datetime(f"{start_year}0101")

    file_list = []
    for var in variables:
        if var not in daily_attrs:
            print(f'no attributes found set for daily "{var}" - skipping')
            continue
        if end_year is None:  # get last date from hourly nc files
            paths = join(ddir_hour, var, f"era5_{var}_*_hourly.nc")
            t1 = get_last_timestep_nc(paths)
        if start_year is None:  # get last date of daily nc files
            paths = join(ddir_day, var, f"era5_{var}_*_daily.nc")
            t0 = get_last_timestep_nc(paths)
            if (t0 > t1) or ((t1 - t0) < pd.to_timedelta(7, unit="d")):
                continue  ## skip if last date within 7 days
            t0 = t0 - dt_era5t
        years = np.unique(pd.date_range(t0, t1, freq="1d").year).tolist()

        # resample to daily values
        for year in years:
            path_resampled = resample_year(
                year, ddir=ddir_hour, outdir=outdir, var=var, dask_kwargs=dask_kwargs
            )
            file_list.extend(path_resampled)

    # write files from outdir to ddir
    if move_to_ddir:
        for filename in file_list:
            var = basename(filename).split("_")[1]
            move_replace(filename, join(ddir_day, var, basename(filename)))


def update_zarr(
    zarr_path: str,
    ddir: str,
    variables: list,
    start_date: str = None,
    end_date: str = None,
    dt_era5t: pd.Timedelta = dt_era5t,
    chunks=None,
    **kwargs,
) -> None:
    """Update zarr file based on nc files.

    Parameters
    ----------
    zarr_path : str
        Zarr store path
    ddir : str
        Root of nc files with path format {ddir}/{var}/eray_{var}_*.nc
    variables : list
        list of variable names to update
    start_date : str, optional
        start date in YYYYMMDD format, by default None and read from last timestep of
        zarr files minus dt_era5t
    end_date : str, optional
        end date in YYYYMMDD format, by default None and read from last timestep of
        the nc files
    chunks: tuple of int
        size of the chunks to store to disk in.
    **kwargs:
        Key-word arguments passed to the zarr driver.
    dt_era5t : pd.Timedelta, optional
        time period of ERA5T data which is overwritten by new data, by default dt_era5t
    """
    print(f"writing to {zarr_path}")
    for var in variables:
        # get start & end dates from files
        paths = join(ddir, var, f"era5_{var}_*.nc")
        if len(glob.glob(paths)) == 0:
            print(f'no nc files found for "{var}" - skipping')
        with xr.open_mfdataset(paths, chunks="auto", mode="r", autoclose=True) as ds_nc:
            t1 = pd.to_datetime(ds_nc["time"][-1].values)
            t0 = pd.to_datetime(ds_nc["time"][0].values)
            if end_date:
                t1 = pd.to_datetime(end_date)
            if start_date:
                t0 = pd.to_datetime(start_date)
            elif os.path.exists(zarr_path):
                with xr.open_zarr(zarr_path, consolidated=False) as ds_zarr:
                    if var in ds_zarr:
                        # check last date with valid values at single location
                        da_zarr = (
                            ds_zarr[var].isel(latitude=0, longitude=0).dropna("time")
                        )
                        t1_zr = pd.to_datetime(da_zarr["time"][-1].values)
                        if t1 > t1_zr:
                            t0 = t1_zr - dt_era5t
                        else:  # already up to date: skip var
                            print(f"{var}: no new data - skipping")
                            continue

        # get zarr chunks to read and process data
        if chunks is None and os.path.exists(zarr_path):
            with xr.open_zarr(zarr_path, consolidated=False) as ds_zarr:
                var0 = var if var in ds_zarr else list(ds_zarr.data_vars.keys())[0]
                chunks = {
                    dim: ds_zarr[var0].chunks[i][0]
                    for i, dim in enumerate(ds_zarr[var0].dims)
                }

        # process year by year a.k.a. file by file
        years = np.unique(pd.date_range(t0, t1, freq="1d").year).tolist()
        for year in years:
            path = glob.glob(join(ddir, var, f"era5_{var}_{year}_*.nc"))[0]
            with xr.open_dataset(
                path, chunks=chunks, mode="r", autoclose=True
            ) as ds_nc:
                ds_nc = ds_nc.sel(time=slice(t0, t1))
                if ds_nc["time"].size > 0:
                    t0_str, t1_str = (
                        ds_nc["time"].dt.strftime("%Y-%m-%d %H:%M").values[[0, -1]]
                    )
                    print(f"{var}: {t0_str} - {t1_str}")
                    # retry at seemingly random Permission Errors ..
                    while True:
                        try:
                            append_zarr(
                                zarr_path,
                                ds_nc,
                                append_dim="time",
                                chunks_out=chunks,
                                **kwargs,
                            )
                            break
                        except PermissionError:
                            print(f"{var}: PermissionError - retry")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Implementaions for downloading and resampling ERA5 data."
    )
    parser.add_argument(
        "--save-zarr", action="store_true", help="Save output to Zarr format"
    )

    args = parser.parse_args()
    os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
    dask_kwargs = {"n_workers": 4, "processes": True}

    save_zarr = False
    outdir = join(r"p:\wflow_global\hydromt_staging\era5")

    # meteo variables
    root = r"p:\wflow_global\hydromt\meteo"
    ddir_hour = join(root, "era5")
    ddir_day = join(root, "era5_daily")
    zarr_hour = join(root, "era5.zarr")
    zarr_day = join(root, "era5_daily.zarr")
    variables_hour = [
        "msl",
        "ssrd",
        "tp",
        "t2m",
        "tisr",
        "tcwv",
        "u10",
        "v10",
        "d2m",
        "ssr",
        "tcc",
        "cape",
    ]
    # NOTE: CAPE excluded from daily values, not sure how it is used and
    # thus how to aggregate
    variables_day = variables_hour[:-1] + ["tmin", "tmax"]

    # hydro / ocean
    # NOTE these are kept in different ddir

    print("downloading ..")
    update_hourly_nc(
        outdir,
        ddir=ddir_hour,
        variables=variables_hour,
        dask_kwargs=dask_kwargs,
        move_to_ddir=True,
    )

    print("resampling to daily values ..")
    update_daily_nc(
        outdir,
        ddir_hour=ddir_hour,
        ddir_day=ddir_day,
        variables=variables_hour,
        dask_kwargs=dask_kwargs,
        move_to_ddir=True,
    )

    if args.save_zarr:
        print("updating hourly zarr..")
        update_zarr(
            zarr_path=zarr_hour,
            ddir=ddir_hour,
            variables=variables_hour,
        )

        print("updating daily zarr..")
        update_zarr(
            zarr_path=zarr_day,
            ddir=ddir_day,
            variables=variables_day,
        )



"""Driver for DataFrames using the pandas library."""

from pathlib import Path
from typing import Any, Callable, Optional

import pandas as pd

from hydromt.data_catalog.drivers.dataframe import DataFrameDriver
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.typing import Variables


class PandasDriver(DataFrameDriver):
    """
    Driver for DataFrames using the pandas library: ``pandas``.

    Supports reading and writing csv, excel (xls, xlsx), parquet, fixed width formatted
    files (fwf) using pandas.
    """

    name = "pandas"
    supports_writing = True

    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        variables: Variables | None = None,
    ) -> pd.DataFrame:
        """
        Read tabular data into a pandas DataFrame using the pandas library.

        Supports multiple file formats including CSV, Parquet, Excel (xls, xlsx),
        and fixed-width text files (fwf). Applies hydromt's `NoDataStrategy` if no
        records are found. Column selection can be controlled through the `variables`
        argument. Only a single file can be read per call.

        Parameters
        ----------
        uris : list[str]
            List containing a single URI to read from. Multiple files are not supported.
        handle_nodata : NoDataStrategy, optional
            Strategy to handle missing or empty data. Default is NoDataStrategy.RAISE.
        variables : Variables | None, optional
            List of columns to read from the file. Ignored if None.

        Returns
        -------
        pd.DataFrame
            The loaded DataFrame containing the data from the source file.

        Raises
        ------
        ValueError
            If multiple files are provided or the file extension is unsupported.
        IOError
            If the file extension is not recognized.
        """
        if len(uris) > 1:
            raise ValueError(
                "DataFrame: Reading multiple files with the "
                f"{self.__class__.__name__} driver is not supported."
            )
        uri = uris[0]
        extension: str = uri.split(".")[-1]

        if extension == "csv":
            variables = self._unify_variables_and_pandas_kwargs(
                uri, pd.read_csv, variables
            )
            df = pd.read_csv(uri, usecols=variables, **self.options.get_kwargs())
        elif extension == "parquet":
            df = pd.read_parquet(uri, columns=variables, **self.options.get_kwargs())
        elif extension in ["xls", "xlsx"]:
            variables = self._unify_variables_and_pandas_kwargs(
                uri, pd.read_excel, variables
            )
            df = pd.read_excel(
                uri,
                usecols=variables,
                engine="openpyxl",
                **self.options.get_kwargs(),
            )
        elif extension in ["fwf", "txt"]:
            df = pd.read_fwf(uri, **self.options.get_kwargs())
        else:
            raise IOError(f"DataFrame: extension {extension} unknown.")

        if df.empty:
            exec_nodata_strat(
                f"No data from {self.name} driver for file uris: {', '.join(uris)}.",
                strategy=handle_nodata,
            )
            return None  # handle_nodata == ignore
        return df

    def write(
        self,
        path: Path | str,
        data: pd.DataFrame,
        *,
        write_kwargs: dict[str, Any] | None = None,
    ) -> Path:
        """
        Write a pandas DataFrame to disk using the pandas library.

        Supports writing to common tabular formats including CSV, Parquet, and Excel (xls, xlsx).
        The file format is automatically inferred from the file extension in the provided path.

        Parameters
        ----------
        path : Path | str
            Destination path where the DataFrame will be saved.
            The file extension determines the output format: `.csv`, `.parquet`, `.xls`, `.xlsx`.
        data : pd.DataFrame
            The DataFrame to be written.
        write_kwargs : dict[str, Any], optional
            Additional keyword arguments passed to the pandas write function (e.g., compression, index, sheet_name).
            Default is None.

        Returns
        -------
        Path
            The path where the DataFrame was written.

        Raises
        ------
        ValueError
            If the path type or file extension is unsupported.
        """
        if isinstance(path, str):
            path = Path(path)
            extension: str = path.suffix[1:]
        elif isinstance(path, Path):
            # suffix includes the '.' which we don't want
            extension: str = path.suffix[1:]
        else:
            raise ValueError(f"unknown pathlike: {path}")

        if extension == "csv":
            data.to_csv(path, **(write_kwargs or {}))
        elif extension == "parquet":
            data.to_parquet(path, **(write_kwargs or {}))
        elif extension in ["xlsx", "xls"]:
            data.to_excel(path, engine="openpyxl", **(write_kwargs or {}))
        else:
            raise ValueError(f"DataFrame: file extension {extension} is unknown.")

        return path

    def _unify_variables_and_pandas_kwargs(
        self,
        uri: str,
        read_method: Callable,
        variables: Optional[Variables],
    ):
        """Prevent clashes between arguments and hydromt query parameters."""
        # include index_col in variables
        if variables:
            if hasattr(self.options, "index_col") and not isinstance(
                self.options.index_col, str
            ):
                # if index_col is an index, get name of col
                new_options: dict[str, Any] = self.options.get_kwargs()
                new_options.pop("index_col", None)
                df: pd.DataFrame = read_method(uri, **{"nrows": 1, **new_options})
                return variables + [df.columns[0]]

        return variables



"""Options for configuring xarray-based drivers in the data catalog."""

import enum
import logging
from os.path import splitext

from pydantic import Field

from hydromt.data_catalog.drivers.base_driver import DriverOptions
from hydromt.data_catalog.drivers.preprocessing import Preprocessor, get_preprocessor

logger = logging.getLogger(__name__)


class XarrayIOFormat(enum.Enum):
    """Enumeration of supported xarray IO formats, each associated with a set of file extensions."""

    ZARR = "zarr"
    NETCDF4 = "netcdf4"

    @property
    def extensions(self) -> set[str]:
        """Get the set of file extensions associated with this IO format."""
        if self == XarrayIOFormat.ZARR:
            return {".zarr"}
        elif self == XarrayIOFormat.NETCDF4:
            return {".nc", ".netcdf"}
        return set()


class XarrayDriverOptions(DriverOptions):
    """Options for configuring xarray-based drivers."""

    preprocess: str | None = Field(
        default=None,
        description="Name of preprocessor to apply before merging datasets. Available preprocessors include: round_latlon, to_datetimeindex, remove_duplicates, harmonise_dims. See their docstrings for details.",
    )
    ext_override: str | None = Field(
        default=None,
        description="Override the file extension check and try to read all files as the given extension. Useful when reading zarr files without the .zarr extension.",
    )

    def get_preprocessor(self) -> Preprocessor:
        """Get the preprocessor instance based on the configured preprocess string."""
        return get_preprocessor(self.preprocess)

    def get_reading_ext(self, uri: str) -> str:
        """Determine the file extension to use for reading, either from the override or from the URI."""
        return self.ext_override or splitext(uri)[-1]

    def get_io_format(self, uri: str) -> XarrayIOFormat | None:
        """Determine the xarray reading format based on the file extension or override.

        Parameters
        ----------
        uri : str
            The URI of the file to read, used to infer the format from the extension if no override is set.

        Returns
        -------
        format : XarrayIOFormat | None
            The xarray reading format, either 'zarr' or 'netcdf4'. Returns None if the extension is unsupported.
        """
        ext = self.get_reading_ext(uri)

        if ext in XarrayIOFormat.ZARR.extensions:
            return XarrayIOFormat.ZARR
        elif ext in XarrayIOFormat.NETCDF4.extensions:
            return XarrayIOFormat.NETCDF4
        return None

    def filter_uris_by_format(
        self, uris: list[str]
    ) -> tuple[list[str], XarrayIOFormat]:
        """Filter the list of URIs to only include those matching the specified format.

        Parameters
        ----------
        uris : list[str]
            List of URIs to filter.

        Returns
        -------
        filtered_uris : list[str]
            List of URIs that match the specified format.
        io_format : XarrayIOFormat
            The xarray reading format to filter by, either 'zarr' or 'netcdf4'.
        """
        io_format = self.get_io_format(uris[0])
        if io_format is None:
            raise ValueError(
                f"Unknown extension for XarrayDriverOptions: {self.get_reading_ext(uris[0])}. "
                "This can be overridden by setting the `ext_override` option to a supported extension (e.g. '.zarr' or '.nc')"
            )

        filtered_uris = []
        for _uri in uris:
            _format = self.get_io_format(_uri)
            if _format is None:
                logger.warning(
                    f"URI {_uri} has unsupported extension for format {io_format.value} and ext_override is not set, skipping..."
                )
            elif _format != io_format:
                logger.warning(
                    f"Reading {io_format.value} and {_uri} has a different format {_format.value}, skipping..."
                )
            else:
                filtered_uris.append(_uri)
        return filtered_uris, io_format



"""Data adapter for DataFrames."""

import logging
from typing import Optional

import numpy as np
import pandas as pd

from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.typing import (
    SourceMetadata,
    TimeRange,
    Variables,
)

logger = logging.getLogger(__name__)


class DataFrameAdapter(DataAdapterBase):
    """Data adapter for DataFrames."""

    def transform(
        self,
        df: pd.DataFrame,
        metadata: Optional[SourceMetadata] = None,
        *,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> pd.DataFrame:
        """Transform data to HydroMT standards.

        Parameters
        ----------
        df : pd.DataFrame
            input DataFrame
        metadata : Optional[SourceMetadata], optional
            MetaData of the source, by default None
        variables : Optional[Variables], optional
            filter for variables, by default None
        time_range : Optional[TimeRange], optional
            filter for start and end times, by default None
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE

        Returns
        -------
        pd.DataFrame
            filtered and harmonized DataFrame

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        # rename variables and parse nodata
        df = self._rename_vars(df)
        df = self._set_nodata(df, metadata)
        # slice data
        df = DataFrameAdapter._slice_data(
            df, variables, time_range, handle_nodata=handle_nodata
        )
        if df is None:
            return None
        # uniformize data
        df = self._apply_unit_conversion(df)
        df = self._set_metadata(df, metadata)
        return df

    def _rename_vars(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.rename:
            rename = {k: v for k, v in self.rename.items() if k in df.columns}
            df = df.rename(columns=rename)
        return df

    def _set_nodata(self, df: pd.DataFrame, metadata: "SourceMetadata") -> pd.DataFrame:
        # parse nodata values
        cols = df.select_dtypes([np.number]).columns
        if metadata.nodata is not None and len(cols) > 0:
            if not isinstance(metadata.nodata, dict):
                nodata = {c: metadata.nodata for c in cols}
            else:
                nodata = metadata.nodata
            for c in cols:
                mv = nodata.get(c, None)
                if mv is not None:
                    is_nodata = np.isin(df[c], np.atleast_1d(mv))
                    df[c] = np.where(is_nodata, np.nan, df[c])
        return df

    @staticmethod
    def _slice_data(
        df: pd.DataFrame,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[pd.DataFrame]:
        """Filter the DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            input DataFrame
        variables : Optional[Variables], optional
            variables to include, or all if None, by default None
        time_range : Optional[TimeRange], optional
            start and end times to include, or all if None, by default None
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE

        Returns
        -------
        Optional[pd.DataFrame]
            filtered DataFrame

        Raises
        ------
        ValueError
            if a variable filter is given which does not match the variables available
        NoDataException
            if no data remains after slicing and handle_nodata == NoDataStrategy.RAISE
        """
        if variables is not None:
            variables = np.atleast_1d(variables).tolist()
            if np.any([var not in df.columns for var in variables]):
                raise ValueError(f"DataFrame: Not all variables found: {variables}")
            df = df.loc[:, variables]

        if time_range is not None and np.dtype(df.index).type == np.datetime64:
            logger.debug(f"Slicing time dime {time_range}")
            idx = df.index.slice_indexer(time_range.start, time_range.end)
            df = df.iloc[idx]
            if df.empty:
                exec_nodata_strat(
                    "DataFrame has no data after time slicing.", handle_nodata
                )
                return None

        return df

    def _apply_unit_conversion(self, df: pd.DataFrame) -> pd.DataFrame:
        unit_names = list(self.unit_mult.keys()) + list(self.unit_add.keys())
        unit_names = [k for k in unit_names if k in df.columns]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} columns.")
        for name in list(set(unit_names)):  # unique
            m = self.unit_mult.get(name, 1)
            a = self.unit_add.get(name, 0)
            df.loc[:, name] = df.loc[:, name] * m + a
        return df

    def _set_metadata(
        self, df: pd.DataFrame, metadata: "SourceMetadata"
    ) -> pd.DataFrame:
        # .`attrs` is not a stable part of the (geo)pandas API, see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.attrs.html#pandas.DataFrame.attrs
        # There are some issues and bug reports from its usage that are very similar to this.
        # (Geo)Pandas' current recommendation is to avoid setting attrs on series within a (Geo)DataFrame and only have one global attrs dict per (Geo)DataFrame.
        # https://github.com/pandas-dev/pandas/issues/35425
        # https://github.com/dask/dask/pull/6742
        # https://github.com/pandas-dev/pandas/issues/41572
        # https://stackoverflow.com/questions/72803056/why-do-attrs-of-series-in-a-pandas-dataframes-change-to-the-dataframe-attrs
        meta = metadata.model_dump()
        attrs = meta.pop("attrs", {})
        for k, v in attrs.items():
            meta[k] = v
        df.attrs.update(**meta)

        return df



"""Spatial Xarrays component."""

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import xarray as xr
from geopandas import GeoDataFrame
from pandas import DataFrame
from xarray import DataArray, Dataset

from hydromt.model.components.base import ModelComponent
from hydromt.model.components.spatial import SpatialModelComponent
from hydromt.model.steps import hydromt_step
from hydromt.readers import open_ncs
from hydromt.typing.type_def import XArrayDict
from hydromt.writers import write_nc

if TYPE_CHECKING:
    from hydromt.model.model import Model

logger = logging.getLogger(__name__)


class SpatialDatasetsComponent(SpatialModelComponent):
    """
    A component to manage collection of geospatial xarray objects.

    It contains a dictionary of xarray DataArray or Dataset objects.
    Compared to ``DatasetsComponent`` this component has a region property.
    """

    def __init__(
        self,
        model: "Model",
        *,
        region_component: str,
        filename: str = "spatial_datasets/{name}.nc",
        region_filename: str = "spatial_datasets/spatial_datasets_region.geojson",
    ):
        """Initialize a SpatialDatasetsComponent.

        Parameters
        ----------
        model: Model
            HydroMT model instance
        region_component: str
            The name of the region component to use as reference for this component's
            region.
        filename: str
            The path to use for reading and writing of component data by default.
            by default ``spatial_datasets/{name}.nc`` ie one file per xarray object in the
            data dictionary.
        region_filename: str
            The path to use for writing the region data to a file. By default
            ``spatial_datasets/spatial_datasets_region.geojson``.
        """
        self._data: XArrayDict | None = None
        self._filename: str = filename
        super().__init__(
            model=model,
            region_component=region_component,
            region_filename=region_filename,
        )

    @property
    def data(self) -> XArrayDict:
        """Model data in the form of xarray objects.

        Return dict of xarray.Dataset or xarray.DataArray objects
        """
        if self._data is None:
            self._initialize()

        assert self._data is not None
        return self._data

    def _initialize(self, skip_read=False) -> None:
        """Initialize geoms."""
        if self._data is None:
            self._data = {}
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    @property
    def _region_data(self) -> GeoDataFrame | None:
        raise AttributeError(
            "region cannot be found in spatialdatasets component."
            "Meaning that the region_component is not set or could not be found in"
            " the model."
        )

    # @hydromt_step
    def set(
        self,
        data: Dataset | DataArray,
        name: str | None = None,
        split_dataset: bool = False,
    ):
        """Add data to the xarray component.

        Arguments
        ---------
        data: xarray.Dataset or xarray.DataArray
            New xarray object to add
        name: str | None, optional
            name of the xarray.
        """
        self._initialize()
        if isinstance(data, DataFrame):
            data = data.to_xarray()
        assert self._data is not None
        if split_dataset:
            if isinstance(data, Dataset):
                ds = {str(name): data[name] for name in data.data_vars}
            else:
                raise ValueError(
                    f"Can only split dataset for Datasets not for {type(data).__name__}"
                )
        elif name is not None:
            ds = {name: data}
        elif data.name is not None:
            ds = {str(data.name): data}
        else:
            raise ValueError(
                "Either name should be set, the data needs to have a name or split_dataset should be True"
            )

        ds = cast(XArrayDict, ds)

        for name, d in ds.items():
            if name in self._data:
                logger.warning(f"Replacing xarray: {name}")
            self._data[name] = d

    @hydromt_step
    def read(self, filename: str | None = None, **kwargs) -> None:
        """Read model dataset files at <root>/<filename>.

        key-word arguments are passed to :py:func:`hydromt.readers.open_nc`

        Parameters
        ----------
        filename : str | None, optional
            Filename relative to model root. Should contain either a * wildcard character
            or a {name} placeholder to read multiple files into the data dictionary.
            All files matching the glob pattern defined by filename will be read.
            If a {name} placeholder is used, that name will be used as the key in the
            data dictionary.
            If no {name} placeholder is used, the filename without extension will be used
            as the key in the data dictionary.
            If None, the path that was provided at init will be used.
        **kwargs:
            Additional keyword arguments that are passed to the
            `hydromt.readers.open_ncs` function.
        """
        self.root._assert_read_mode()
        self._initialize(skip_read=True)
        kwargs = {**{"engine": "netcdf4"}, **kwargs}
        placeholder_filename = str(filename or self._filename).replace("*", "{name}")
        for name, ds in open_ncs(
            placeholder_filename, root=self.root.path, **kwargs
        ).items():
            self._open_datasets.append(ds)
            if len(ds.data_vars) == 1:
                (da,) = ds.data_vars.values()
            else:
                da = ds
            self.set(data=da, name=name)

    @hydromt_step
    def write(
        self,
        filename: str | None = None,
        *,
        gdal_compliant: bool = False,
        rename_dims: bool = False,
        force_sn: bool = False,
        to_netcdf_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Write dictionary of xarray.Dataset and/or xarray.DataArray to netcdf files.

        Possibility to update the xarray objects attributes to get GDAL compliant NetCDF
        files, using :py:meth:`~hydromt.raster.gdal_compliant`.
        The function will first try to directly write to file. In case of
        PermissionError, it will first write a temporary file and then move the file.

        key-word arguments are passed to :py:meth:`xarray.Dataset.to_netcdf`

        Parameters
        ----------
        nc_dict: dict[str, xr.Dataset | xr.DataArray]
            Dictionary of xarray.Dataset and/or xarray.DataArray to write
        fn: str
            filename relative to model root and should contain a {name} placeholder
        gdal_compliant: bool, optional
            If True, convert xarray.Dataset and/or xarray.DataArray to gdal compliant
            format using :py:meth:`~hydromt.raster.gdal_compliant`
        rename_dims: bool, optional
            If True, rename x_dim and y_dim to standard names depending on the CRS
            (x/y for projected and lat/lon for geographic). Only used if
            ``gdal_compliant`` is set to True. By default, False.
        force_sn: bool, optional
            If True, forces the dataset to have South -> North orientation. Only used
            if ``gdal_compliant`` is set to True. By default, False.
        to_netcdf_kwargs: dict[str, Any] | None, optional
            Additional keyword arguments that are passed to the `to_netcdf`
            function.
        """
        self.root._assert_write_mode()

        if len(self.data) == 0:
            logger.info(
                f"{self.model.name}.{self.name_in_model}: No data found, skipping writing."
            )
            return

        to_netcdf_kwargs = to_netcdf_kwargs or {}
        to_netcdf_kwargs.setdefault("engine", "netcdf4")
        placeholder_filename = str(filename or self._filename).replace("*", "{name}")
        for name, ds in self.data.items():
            file_path = self.root.path / placeholder_filename.format(name=name)
            file_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(
                f"{self.model.name}.{self.name_in_model}: Writing spatial dataset to {file_path}."
            )

            close_handle = write_nc(
                ds,
                file_path=file_path,
                gdal_compliant=gdal_compliant,
                rename_dims=rename_dims,
                force_sn=force_sn,
                force_overwrite=self.root.mode.is_override_mode(),
                to_netcdf_kwargs=to_netcdf_kwargs,
            )
            if close_handle is not None:
                self._deferred_file_close_handles.append(close_handle)

    @hydromt_step
    def add_raster_data_from_rasterdataset(
        self,
        raster_filename: str | Path | Dataset,
        variables: list[str] | None = None,
        fill_method: str | None = None,
        name: str | None = None,
        reproject_method: str | None = None,
        split_dataset: bool = True,
        rename: dict[str, str] | None = None,
    ) -> list[str]:
        """HYDROMT CORE METHOD: Add data variable(s) from ``raster_filename`` to datasets component.

        If raster is a xarray dataset, all variables will be added unless ``variables``
        list is specified.

        Adds model layers:
        * **raster.name**: data from raster_filename

        Parameters
        ----------
        raster_filename: str, Path, xr.Dataset
            Data catalog key, path to raster file or raster xarray data object.
        variables: list, optional
            list of variables to add to datasets from raster_filename. By default all.
        fill_method : str, optional
            If specified, fills nodata values using fill_nodata method.
            Available methods are {'linear', 'nearest', 'cubic', 'rio_idw'}.
        name: str, optional
            Name of new dataset in self.data dictionary,
            only in case split_dataset=False.
        reproject_method: str, optional
            See rasterio.warp.reproject for existing methods, by default the data is
            not reprojected (None).
        split_dataset: bool, optional
            If data is a xarray.Dataset split it into several xarray.DataArrays.
        rename: dict, optional
            Dictionary to rename variable names in raster_filename before adding to the datasets
            {'name_in_raster_filename': 'name_in_dataset'}. By default empty.

        Returns
        -------
        list
            Names of added model map layers
        """
        rename = rename or {}
        logger.info(f"Preparing dataset data from raster source {raster_filename}")
        # Read raster data and select variables
        ds = self.data_catalog.get_rasterdataset(
            raster_filename,
            geom=self.region,
            buffer=2,
            variables=variables,
            single_var_as_array=False,
        )
        if isinstance(ds, DataArray):
            ds = ds.to_dataset()

        ds = cast(Dataset, ds)
        # Fill nodata
        if fill_method is not None:
            ds = ds.raster.interpolate_na(method=fill_method)
        # Reprojection
        if ds.rio.crs != self.model.crs and reproject_method is not None:
            ds = ds.raster.reproject(dst_crs=self.model.crs, method=reproject_method)
        self.set(ds.rename(rename), name=name, split_dataset=split_dataset)

        return list(ds.data_vars.keys())

    @hydromt_step
    def add_raster_data_from_raster_reclass(
        self,
        raster_filename: str | Path | DataArray,
        reclass_table_filename: str | Path | DataFrame,
        reclass_variables: list[str],
        variable: str | None = None,
        fill_method: str | None = None,
        reproject_method: str | None = None,
        name: str | None = None,
        split_dataset: bool = True,
        rename: dict[str, str] | None = None,
        **kwargs,
    ) -> list[str]:
        r"""HYDROMT CORE METHOD: Add data variable(s) to datasets component by reclassifying the data in ``raster_filename`` based on ``reclass_table_filename``.

        This is done by reclassifying the data in
        ``raster_filename`` based on ``reclass_table_filename``.

        Adds model layers:

        * **reclass_variables**: reclassified raster data

        Parameters
        ----------
        raster_filename: str, Path, xr.DataArray
            Data catalog key, path to raster file or raster xarray data object.
            Should be a DataArray. Else use `variable` argument for selection.
        reclass_table_filename: str, Path, pd.DataFrame
            Data catalog key, path to tabular data file or tabular pandas dataframe
            object for the reclassification table of `raster_filename`.
        reclass_variables: list[str]
            list of reclass_variables from reclass_table_filename table to add to the datasets. Index
            column should match values in `raster_filename`.
        variable: str | None = None
            Name of raster dataset variable to use. This is only required when reading
            datasets with multiple variables. By default None.
        fill_method : str, optional
            If specified, fills nodata values in `raster_filename` using fill_nodata method
            before reclassifying. Available methods are {'linear', 'nearest',
            'cubic', 'rio_idw'}.
        reproject_method: str, optional
            See rasterio.warp.reproject for existing methods, by default the data is
            not reprojected (None).
        name: str, optional
            Name of new dataset variable, only in case split_dataset=False.
        split_dataset: bool, optional
            If data is a xarray.Dataset split it into several xarray.DataArrays.
        rename: dict, optional
            Dictionary to rename variable names in reclass_variables before adding to
            grid {'name_in_reclass_table': 'name_in_grid'}. By default empty.
        \**kwargs:
            Additional keyword arguments that are passed to the
            `data_catalog.get_rasterdataset` function.

        Returns
        -------
        list
            Names of added model map layers
        """  # noqa: E501
        rename = rename or {}
        logger.info(
            f"Preparing map data by reclassifying the data in {raster_filename} based"
            f" on {reclass_table_filename}"
        )
        # Read raster data and remapping table
        da = self.data_catalog.get_rasterdataset(
            raster_filename,
            geom=self.region,
            buffer=2,
            variables=variable,
            **kwargs,
        )
        if not isinstance(da, DataArray):
            raise ValueError(
                f"raster_filename {raster_filename} should be a single variable. "
                "Please select one using the 'variable' argument"
            )
        df_vars = self.data_catalog.get_dataframe(
            reclass_table_filename, variables=reclass_variables
        )
        # Fill nodata
        if fill_method is not None:
            da = da.raster.interpolate_na(method=fill_method)
        # Mapping function
        ds_vars = da.raster.reclassify(reclass_table=df_vars, method="exact")
        # Reprojection
        if ds_vars.rio.crs != self.model.crs and reproject_method is not None:
            ds_vars = ds_vars.raster.reproject(dst_crs=self.model.crs)
        self.set(ds_vars.rename(rename), name=name, split_dataset=split_dataset)

        return list(ds_vars.data_vars.keys())

    def test_equal(self, other: ModelComponent) -> tuple[bool, dict[str, str]]:
        """
        Test if two DatasetsComponents are equal.

        Parameters
        ----------
        other: ModelComponent
            The other ModelComponent to compare with.

        Returns
        -------
        tuple[bool, dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_datasets = cast(SpatialDatasetsComponent, other)
        for name, ds in self.data.items():
            if name not in other_datasets.data:
                errors[name] = "Dataset not found in other component."
            try:
                xr.testing.assert_allclose(ds, other_datasets.data[name])
            except AssertionError as e:
                errors[name] = str(e)

        return len(errors) == 0, errors



"""Pydantic models for validation of various hydromt internal components."""

from enum import Enum

from hydromt._validators.data_catalog_v0x import (
    DataCatalogV0Item,
    DataCatalogV0ItemMetadata,
    DataCatalogV0MetaData,
    DataCatalogV0Validator,
)
from hydromt._validators.data_catalog_v1x import (
    DataCatalogV1Item,
    DataCatalogV1ItemMetadata,
    DataCatalogV1MetaData,
    DataCatalogV1Validator,
)
from hydromt._validators.model_config import HydromtModelSetup, HydromtModelStep
from hydromt._validators.region import (
    BoundingBoxRegion,
    PathRegion,
    Region,
    validate_region,
)

__all__ = [
    "DataCatalogV1Item",
    "DataCatalogV1ItemMetadata",
    "DataCatalogV1MetaData",
    "DataCatalogV1Validator",
    "DataCatalogV0Item",
    "DataCatalogV0ItemMetadata",
    "DataCatalogV0MetaData",
    "DataCatalogV0Validator",
    "BoundingBoxRegion",
    "PathRegion",
    "Region",
    "validate_region",
    "HydromtModelStep",
    "HydromtModelSetup",
    "Format",
]


class Format(Enum):
    v0 = 0
    v1 = 1

    # just a convenience function
    # with some nicer error messages
    @classmethod
    def from_str(cls, s: str) -> "Format":
        try:
            return cls[s.strip().lower()]
        except KeyError as e:
            raise ValueError(
                f"{e} is not a known valid Format, options are {list(cls.__members__.keys())}"
            )



"""DatasetDriver for zarr data."""

import logging
from pathlib import Path
from typing import Any, ClassVar

import xarray as xr
from pydantic import Field

from hydromt.data_catalog.drivers.base_driver import (
    DRIVER_OPTIONS_DESCRIPTION,
)
from hydromt.data_catalog.drivers.dataset.dataset_driver import DatasetDriver
from hydromt.data_catalog.drivers.xarray_options import (
    XarrayDriverOptions,
    XarrayIOFormat,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat

logger = logging.getLogger(__name__)


class DatasetXarrayDriver(DatasetDriver):
    """
    Driver for Dataset using xarray: ``dataset_xarray``.

    Supports reading and writing zarr and netcdf files using xarray.
    zarr files will be read using `xr.open_zarr` and netcdf files using
    `xr.open_mfdataset`.

    """

    name: ClassVar[str] = "dataset_xarray"
    supports_writing: ClassVar[bool] = True
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = (
        XarrayIOFormat.ZARR.extensions | XarrayIOFormat.NETCDF4.extensions
    )

    options: XarrayDriverOptions = Field(
        default_factory=XarrayDriverOptions, description=DRIVER_OPTIONS_DESCRIPTION
    )

    def read(
        self, uris: list[str], *, handle_nodata: NoDataStrategy = NoDataStrategy.RAISE
    ) -> xr.Dataset:
        """
        Read zarr or netCDF data into an xarray Dataset.

        Supports reading multiple compatible datasets and merging them into a single
        xarray Dataset. File format is automatically inferred from the file extension,
        unless overridden via the driver options. Optionally applies a preprocessor
        function to each dataset before merging.

        Parameters
        ----------
        uris : list[str]
            List of URIs to read data from. All files must share the same format.
        handle_nodata : NoDataStrategy, optional
            Strategy to handle missing or empty data. Default is NoDataStrategy.RAISE.

        Returns
        -------
        xr.Dataset
            The dataset read from the source files.

        Raises
        ------
        ValueError
            If the provided files have mixed or unsupported extensions.
        """
        preprocessor = self.options.get_preprocessor()
        filtered_uris, io_format = self.options.filter_uris_by_format(uris)

        # Read and merge
        if io_format == XarrayIOFormat.ZARR:
            datasets = [
                preprocessor(xr.open_zarr(_uri, **self.options.get_kwargs()))
                for _uri in filtered_uris
            ]
            ds: xr.Dataset = xr.merge(datasets)
        elif io_format == XarrayIOFormat.NETCDF4:
            ds: xr.Dataset = xr.open_mfdataset(
                filtered_uris,
                decode_coords="all",
                preprocess=preprocessor,
                **self.options.get_kwargs(),
                decode_timedelta=True,
            )
        else:
            raise ValueError(
                f"Unknown extension for DatasetXarrayDriver: {self.options.get_reading_ext(uris[0])}"
            )

        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
                return None  # handle_nodata == ignore
        return ds

    def write(
        self,
        path: Path | str,
        data: xr.Dataset,
        *,
        write_kwargs: dict[str, Any] | None = None,
    ) -> Path:
        """
        Write an xarray Dataset to disk using the xarray I/O engine.

        Supports writing to both Zarr and NetCDF formats. The file format is inferred
        from the file extension. If the extension is not recognized, a ValueError is raised.

        Parameters
        ----------
        path : Path | str
            Destination path or URI where the Dataset will be written.
            The file extension determines the output format:
            `.zarr`, `.nc`, or `.netcdf`.
        data : xr.Dataset
            The Dataset to write to disk.
        write_kwargs : dict[str, Any], optional
            Additional keyword arguments passed to the xarray write function
            (`Dataset.to_zarr` or `Dataset.to_netcdf`). Default is None.

        Returns
        -------
        Path
            The path where the dataset was written.

        Raises
        ------
        ValueError
            If the provided file extension is unsupported.
        """
        fmt = self.options.get_io_format(path)
        write_kwargs = write_kwargs or {}
        if fmt == XarrayIOFormat.ZARR:
            write_kwargs.setdefault("zarr_format", 2)
            data.to_zarr(path, **write_kwargs)
        elif fmt == XarrayIOFormat.NETCDF4:
            data.to_netcdf(path, **write_kwargs)
        else:
            raise ValueError(
                f"Unknown extension for DatasetXarrayDriver: {self.options.get_reading_ext(path)}"
            )

        return path



"""All of the types for handeling errors within HydroMT."""

import inspect
import logging
from enum import Enum


class DeprecatedError(Exception):
    """Simple custom class to raise an error for something that is now deprecated."""

    def __init__(self, msg: str):
        """Initialise the object."""
        self.base = "DeprecationError"
        self.message = msg

    def __str__(self):
        return f"{self.base}: {self.message}"


class NoDataStrategy(Enum):
    """Strategy to handle nodata values."""

    RAISE = "raise"
    WARN = "warn"
    IGNORE = "ignore"


class NoDataException(Exception):
    """Exception raised for errors in the input.

    Attributes
    ----------
        message -- explanation of the error
    """

    def __init__(self, message="No data available"):
        self.message = message
        super().__init__(self.message)


def exec_nodata_strat(msg: str, strategy: NoDataStrategy) -> None:
    """Execute nodata strategy.

    Uses the logger from the calling module if it has a logger.
    Otherwise creates a new logger with the calling module's name.
    Otherwise uses a backup logger from this current module.
    """
    frame = inspect.currentframe()
    caller_frame = frame.f_back if frame else None
    module = inspect.getmodule(caller_frame) if caller_frame else None
    logger_name = getattr(module, "__name__", __name__)
    logger = getattr(module, "logger", logging.getLogger(logger_name))

    if strategy == NoDataStrategy.RAISE:
        raise NoDataException(msg)
    elif strategy == NoDataStrategy.WARN:
        logger.warning(msg)
    elif strategy == NoDataStrategy.IGNORE:
        # do nothing
        pass



"""Accessors to extend the functionality of xarray structures."""

# required for accessor style documentation
from xarray import DataArray, Dataset  # noqa: F401

from hydromt.gis import flw
from hydromt.gis.gis_utils import parse_crs, utm_crs, zoom_to_overview_level
from hydromt.gis.raster import RasterDataArray, RasterDataset
from hydromt.gis.raster_utils import (
    cellres,
    full,
    full_from_transform,
    full_like,
    merge,
    spread2d,
)
from hydromt.gis.vector import GeoDataArray, GeoDataset
from hydromt.gis.vector_utils import nearest, nearest_merge

__all__ = [
    "RasterDataArray",
    "RasterDataset",
    "GeoDataArray",
    "GeoDataset",
    "cellres",
    "full",
    "full_from_transform",
    "full_like",
    "merge",
    "nearest_merge",
    "nearest",
    "spread2d",
    "utm_crs",
    "parse_crs",
    "zoom_to_overview_level",
    "flw",
]



"""Base class for different drivers."""

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, ClassVar

from pydantic import BaseModel, ConfigDict, Field, field_validator

from hydromt._abstract_base import AbstractBaseModel
from hydromt.error import NoDataStrategy
from hydromt.plugins import PLUGINS
from hydromt.typing.fsspec_types import (
    FSSpecFileSystem,
)

logger = logging.getLogger(__name__)

DRIVER_OPTIONS_DESCRIPTION = """
Driver options that can be used to configure the behavior of the driver.
DriverOptions allows for setting arbitrary kwargs. Any options not explicitly declared
in the DriverOptions class are passed as kwargs to the underlying `open` functions.
"""


class DriverOptions(BaseModel):
    """Options for the driver.

    Fields declared in this class or its subclasses are used to configure the behavior of the driver.
    Any other undeclared keyword arguments are allowed and will be stored in the instance.
    These extra keyword arguments are passed as kwargs to the underlying `open` functions used by the driver.

    Retrieving the extra kwargs can be done using the `get_kwargs` method:
    - Some options may be solely used internally by the driver and not passed to the `open` functions.
    - Some options may be used both internally and passed to the `open` functions.
    - To specify which declared fields should be passed to the `open` functions, set the class variable `KWARGS_FOR_OPEN`.
    - By default, all declared fields are excluded, and only extra kwargs are passed.

    """

    # allow arbitrary kwargs, usually passed to some `open_dataset` function()
    model_config = ConfigDict(extra="allow")

    KWARGS_FOR_OPEN: ClassVar[set[str]] = set()
    """Fields that are to be included as kwargs when passing to `open` functions.
    By default, all fields defined in the class are excluded."""

    @classmethod
    def get_excluded_kwarg_names_for_open(cls) -> set[str]:
        """Get the fields that are to be excluded when passing to `open` functions.

        This includes all declared fields in the class, minus those defined in `KWARGS_FOR_OPEN`.
        """
        return set(cls.model_fields.keys()).difference(cls.KWARGS_FOR_OPEN)

    def to_dict(self, exclude: set[str] | None = None) -> dict:
        """Return dict of kwargs excluding reserved/internal keys, and including the extra kwargs."""
        return self.model_dump(exclude=exclude, exclude_unset=True)

    def get_kwargs(self) -> dict[str, Any]:
        """Return attributes set that are not explicitly declared fields."""
        return self.model_dump(
            exclude=self.get_excluded_kwarg_names_for_open(), exclude_unset=True
        )


class BaseDriver(AbstractBaseModel, ABC):
    """Base class for different drivers.

    Is used to implement common functionality.
    """

    supports_writing: ClassVar[bool] = False
    """"Whether the driver supports writing data."""

    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")
    """"Forbid extra fields in the driver configuration and allow arbitrary types."""

    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = set()
    """"Set of supported file extensions for this driver."""

    filesystem: FSSpecFileSystem = Field(
        default_factory=FSSpecFileSystem,
        description="Filesystem to use for accessing the data.",
    )
    options: DriverOptions = Field(
        default_factory=DriverOptions, description=DRIVER_OPTIONS_DESCRIPTION
    )

    def __eq__(self, value: object) -> bool:
        """Compare equality.

        Overwritten as filesystem between two sources can not be the same.
        """
        if isinstance(value, self.__class__):
            return self.model_dump(round_trip=True) == value.model_dump(round_trip=True)
        return super().__eq__(value)

    @classmethod
    def _load_plugins(cls):
        """Load Driver plugins."""
        plugins = PLUGINS.driver_plugins.keys()
        logger.debug(f"loaded {cls.__name__} plugins: {', '.join(plugins)}")

    @abstractmethod
    def read(
        self, uris: list[str], *, handle_nodata: NoDataStrategy = NoDataStrategy.RAISE
    ) -> Any:
        """Read data using the driver.

        Parameters
        ----------
        uris : list[str]
            List of URIs to read data from.
        handle_nodata : NoDataStrategy, optional
            Strategy to handle no data situations. Default is NoDataStrategy.RAISE.
        """
        ...

    @abstractmethod
    def write(
        self,
        path: Path | str,
        data: Any,
        *,
        write_kwargs: dict[str, Any] | None = None,
    ) -> Path:
        """
        Write data using the driver.

        Parameters
        ----------
        path : Path | str
            Destination path where the data will be written. Can be a local or remote URI.
        data : Any
            The data object to write. The expected type depends on the specific driver implementation
            (e.g., pandas DataFrame, xarray Dataset, GeoDataFrame, etc.).
        write_kwargs : dict[str, Any], optional
            Additional keyword arguments passed to the underlying writer function (e.g., `to_csv`, `to_zarr`, etc.).
            Default is None.

        Returns
        -------
        Path
            The path where the data was written.
        """
        ...

    @field_validator("filesystem", mode="before")
    @classmethod
    def _create_filesystem(cls, v: Any) -> FSSpecFileSystem:
        """Allow filesystem to be provided as string, dict, or FSSpecFileSystem."""
        return FSSpecFileSystem.create(v)



"""Implementation for the RasterDatasetAdapter."""

from __future__ import annotations

import logging
from typing import Optional

import numpy as np
import xarray as xr

from hydromt._utils import (
    _has_no_data,
    _set_metadata,
    _shift_dataset_time,
    _single_var_as_array,
)
from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.typing import (
    Data,
    SourceMetadata,
    TimeRange,
    Variables,
)

logger = logging.getLogger(__name__)

__all__ = ["DatasetAdapter"]


class DatasetAdapter(DataAdapterBase):
    """Implementation for the DatasetAdapter."""

    def transform(
        self,
        ds: xr.Dataset,
        metadata: SourceMetadata,
        *,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        single_var_as_array: bool = True,
    ) -> Optional[xr.Dataset]:
        """Return a clipped, sliced and unified Dataset.

        For a detailed description see:
        :py:func:`~hydromt.data_catalog.DataCatalog.get_dataset`
        """
        # rename variables and parse data and attrs
        ds = self._rename_vars(ds)
        ds = self._set_nodata(ds, metadata)
        ds = self._shift_time(ds)
        # slice
        ds = DatasetAdapter._slice_data(
            ds, variables, time_range, handle_nodata=handle_nodata
        )
        if ds is None:
            return None
        # uniformize
        ds = self._apply_unit_conversion(ds)
        ds = _set_metadata(ds, metadata)
        # return array if single var and single_var_as_array
        ds = _single_var_as_array(ds, single_var_as_array, variable_name=variables)
        return ds

    def _rename_vars(self, ds: Data) -> Data:
        rm = {k: v for k, v in self.rename.items() if k in ds}
        ds = ds.rename(rm)
        return ds

    def _set_nodata(self, ds: Data, metadata: SourceMetadata) -> Data:
        if metadata.nodata is not None:
            if not isinstance(metadata.nodata, dict):
                nodata = {k: metadata.nodata for k in ds.data_vars.keys()}
            else:
                nodata = metadata.nodata
            for k in ds.data_vars:
                mv = nodata.get(k, None)
                if mv is not None and ds[k].attrs.get("_FillValue", None) is None:
                    ds[k].attrs["_FillValue"] = mv
        return ds

    def _apply_unit_conversion(self, ds: Data) -> Data:
        unit_names = list(self.unit_mult.keys()) + list(self.unit_add.keys())
        unit_names = [k for k in unit_names if k in ds.data_vars]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} variables.")
        for name in list(set(unit_names)):  # unique
            m = self.unit_mult.get(name, 1)
            a = self.unit_add.get(name, 0)
            da = ds[name]
            attrs = da.attrs.copy()
            nodata_isnan = da.attrs.get("_FillValue", None) is None or np.isnan(
                da.attrs.get("_FillValue", None)
            )
            # nodata value is explicitly set to NaN in case no nodata value is provided
            nodata = np.nan if nodata_isnan else da.attrs["_FillValue"]
            data_bool = ~np.isnan(da) if nodata_isnan else da != nodata
            ds[name] = xr.where(data_bool, da * m + a, nodata)
            ds[name].attrs.update(attrs)  # set original attributes
        return ds

    def _shift_time(self, ds: Data) -> Data:
        dt = self.unit_add.get("time", 0)
        return _shift_dataset_time(dt=dt, ds=ds)

    @staticmethod
    def _slice_data(
        ds: Data,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[Data]:
        """Slice the dataset in space and time.

        Arguments
        ---------
        ds : xarray.Dataset or xarray.DataArray
            The Dataset to slice.
        variables : str or list of str, optional.
            Names of variables to return.
        time_range: tuple of str, datetime, optional
            Start and end date of period of interest. By default the entire time period
            of the dataset is returned.
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE

        Returns
        -------
        ds : xarray.Dataset
            The sliced Dataset.
        """
        if isinstance(ds, xr.DataArray):
            if ds.name is None:
                ds.name = "data"
            ds = ds.to_dataset()
        elif variables is not None:
            variables = np.atleast_1d(variables).tolist()
            if len(variables) > 1 or len(ds.data_vars) > 1:
                mvars = [var not in ds.data_vars for var in variables]
                if any(mvars):
                    raise ValueError(f"Dataset: variables not found {mvars}")
                ds = ds[variables]

        if time_range is not None:
            ds = DatasetAdapter._slice_temporal_dimension(
                ds, time_range, handle_nodata=handle_nodata
            )
            if ds is None:
                return None

        return ds

    @staticmethod
    def _slice_temporal_dimension(
        ds: Data,
        time_range: TimeRange,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[Data]:
        if (
            "time" in ds.dims
            and ds["time"].size > 1
            and np.issubdtype(ds["time"].dtype, np.datetime64)
        ):
            logger.debug(f"Slicing time dim {time_range}")
            ds = ds.sel(time=slice(time_range.start, time_range.end))
            if _has_no_data(ds):
                exec_nodata_strat("No data left after time slicing.", handle_nodata)
                return None
        return ds



"""Pydantic models for the validation of Data catalogs."""

from logging import warning
from pathlib import Path
from typing import Any, Dict, List, Literal

from packaging.specifiers import SpecifierSet
from packaging.version import Version
from pydantic import (
    BaseModel,
    ConfigDict,
    field_serializer,
    field_validator,
    model_serializer,
    model_validator,
)
from pydantic.fields import Field
from pydantic_core import Url, ValidationError
from pyproj import CRS
from pyproj.exceptions import CRSError

from hydromt import __version__ as HYDROMT_VERSION
from hydromt._validators.data_catalog_v0x import (
    DataCatalogV0Item,
    DataCatalogV0ItemMetadata,
    DataCatalogV0MetaData,
    DataCatalogV0Validator,
)
from hydromt.data_catalog.drivers.base_driver import DriverOptions
from hydromt.readers import _yml_from_uri_or_path
from hydromt.typing import Bbox, Number, TimeRange

DRIVER_RENAME_MAPPING: Dict[str, Dict[str, str]] = {
    "RasterDataset": {
        "raster": "rasterio",
        "zarr": "raster_xarray",
        "netcdf": "raster_xarray",
        "raster_tindex": "rasterio",
    },
    "GeoDataset": {
        "vector": "geodataset_vector",
        "zarr": "geodataset_xarray",
        "netcdf": "geodataset_xarray",
    },
    "GeoDataFrame": {
        "csv": "geodataframe_table",
        "parquet": "geodataframe_table",
        "xls": "geodataframe_table",
        "xlsx": "geodataframe_table",
        "xy": "geodataframe_table",
        "vector": "pyogrio",
    },
    "DataFrame": {
        "csv": "pandas",
        "parquet": "pandas",
        "xls": "pandas",
        "xlsx": "pandas",
        "fwf": "pandas",
    },
    "DataSet": {
        "zarr": "dataset_xarray",
        "netcdf": "dataset_xarray",
    },
}


class SourceSpec(BaseModel):
    """A complete source variant specification."""

    source: str
    provider: str | None = None
    version: str | Number | None = None

    @staticmethod
    def from_dict(input_dict):
        """Create a source variant specification from a dictionary."""
        return SourceSpec(
            source=input_dict["source"],
            provider=input_dict.get("provider", None),
            version=input_dict.get("version", None),
        )


class DataCatalogV1UriResolverItem(BaseModel):
    name: str
    options: Dict[str, Any] | None = None


class DataCatalogV1DriverItem(BaseModel):
    name: str
    options: DriverOptions | None = None


class SourceVariant(BaseModel):
    """A variant for a data source."""

    provider: str | None = None
    version: str | Number | None = None
    uri: Path | None = None
    rename: Dict[str, str] | None = None
    filesystem: Literal["local", "s3", "gcs"] | None = None
    storage_options: Dict[str, Any] | None = None
    driver: DataCatalogV1DriverItem | None = None


class Extent(BaseModel):
    """A validation model for describing the space and time a dataset covers."""

    time_range: TimeRange
    bbox: Bbox


class DataCatalogV1MetaData(BaseModel):
    """The metadata section of a Hydromt data catalog."""

    roots: List[Path] | None = None
    version: str | Number | None = None
    hydromt_version: str | None = None
    name: str | None = None
    category: str | None = None
    validate_hydromt_version: bool = True

    model_config = ConfigDict(
        str_strip_whitespace=True,
        extra="allow",
    )

    @staticmethod
    def from_v0(v0_metadata: DataCatalogV0MetaData, category: str | None = None):
        return DataCatalogV1MetaData(
            roots=v0_metadata.roots,
            version=v0_metadata.version,
            hydromt_version=v0_metadata.hydromt_version,
            name=v0_metadata.name,
            category=category,
            validate_hydromt_version=v0_metadata.validate_hydromt_version,
        )

    @model_validator(mode="after")
    def _check_version_compatible(self) -> "DataCatalogV1MetaData":
        if self.hydromt_version is None:
            warning(
                f"No hydromt version was specified for the data catalog, thus compatibility between used hydromt version ({HYDROMT_VERSION}) and the catalog could not be determined."
            )
            return self
        requested = SpecifierSet(self.hydromt_version, prereleases=True)
        version = Version(HYDROMT_VERSION)

        if version in requested:
            return self
        else:
            raise ValueError(
                f"Current hydromt version {HYDROMT_VERSION} is not compatible with version specified in catalog {self.hydromt_version}"
            )

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "DataCatalogV1MetaData":
        """Convert a dictionary into a validated data catalog metadata item."""
        return DataCatalogV1MetaData(**input_dict)


class DataCatalogV1ItemMetadata(BaseModel):
    """The metadata for a data source."""

    crs: str | int | None = None
    nodata: Number | dict[str, Number] | None = None
    attrs: Dict[str, Any] | None = None
    category: str | None = None
    paper_doi: str | None = None
    paper_ref: str | None = None
    source_license: str | None = None
    source_url: str | None = None
    source_version: str | None = None
    notes: str | None = None
    temporal_extent: dict | None = None
    spatial_extent: dict | None = None

    model_config = ConfigDict(
        str_strip_whitespace=True,
        coerce_numbers_to_str=True,
        extra="allow",
    )

    @field_validator("crs", mode="after")
    @classmethod
    def _check_valid_crs(cls, value):
        try:
            if value:
                _ = CRS.from_user_input(value)
        except CRSError as e:
            raise ValueError(e)
        return value

    @staticmethod
    def from_v0(
        v0_metadata: DataCatalogV0ItemMetadata | None,
        crs: str | int | None = None,
        nodata: Number | dict[str, Number] | None = None,
        attrs: Dict[str, Any] | None = None,
    ):
        if (
            (v0_metadata is None or v0_metadata.is_empty())
            and not crs
            and not nodata
            and not attrs
        ):
            return None
        elif v0_metadata:
            return DataCatalogV1ItemMetadata(
                crs=crs,
                nodata=nodata,
                attrs=attrs,
                category=v0_metadata.category,
                paper_doi=v0_metadata.paper_doi,
                paper_ref=v0_metadata.paper_ref,
                source_license=v0_metadata.source_license,
                source_url=v0_metadata.source_url,
                source_version=v0_metadata.source_version,
                notes=v0_metadata.notes,
                temporal_extent=v0_metadata.temporal_extent,
                spatial_extent=v0_metadata.spatial_extent,
                **v0_metadata.model_extra,
            )
        else:
            return DataCatalogV1ItemMetadata(crs=crs, nodata=nodata, attrs=attrs)

    @staticmethod
    def from_dict(input_dict):
        """Convert a dictionary into a validated source metadata item."""
        if input_dict is None:
            return DataCatalogV1ItemMetadata()
        else:
            item_source_url = input_dict.pop("source_url", None)
            if item_source_url:
                Url(item_source_url)
            return DataCatalogV1ItemMetadata(**input_dict, source_url=item_source_url)


class DataCatalogV1DataAdapter(BaseModel):
    rename: Dict[str, str] | None = None
    unit_mult: Dict[str, int | float] | None = None
    unit_add: Dict[str, int | float] | None = None


class DataCatalogV1Item(BaseModel):
    """A validated data source."""

    # these are required but they can be defined in the variants instead
    driver: DataCatalogV1DriverItem | None = None
    uri: Path | None = None

    uri_resolver: DataCatalogV1UriResolverItem | None = None
    name: str
    data_type: Literal["RasterDataset", "GeoDataset", "GeoDataFrame", "DataFrame"]
    provider: str | None = None
    metadata: DataCatalogV1ItemMetadata | None = None
    variants: List[SourceVariant] | None = None
    version: str | Number | None = None
    placeholder: dict[str, list[str]] | None = None
    data_adapter: DataCatalogV1DataAdapter | None = None

    model_config = ConfigDict(
        str_strip_whitespace=True,
        extra="forbid",
    )

    @field_serializer("uri")
    def serialize_uri(self, value):
        return value.as_posix()

    @staticmethod
    def from_v0(v0_item: DataCatalogV0Item) -> "DataCatalogV1Item":
        uri = v0_item.path if v0_item.path else None
        possible_drivers = DRIVER_RENAME_MAPPING[v0_item.data_type]
        driver_name = possible_drivers.get(v0_item.driver, None)
        options = {}
        if v0_item.driver_kwargs and len(v0_item.driver_kwargs) > 0:
            options = v0_item.driver_kwargs

        if v0_item.kwargs and len(v0_item.kwargs) > 0:
            options = {**options, **v0_item.kwargs}

        if driver_name:
            driver = DataCatalogV1DriverItem(
                name=driver_name, options=options if len(options) > 0 else None
            )
        else:
            driver = None

        metadata = DataCatalogV1ItemMetadata.from_v0(
            v0_metadata=v0_item.meta,
            crs=v0_item.crs,
            nodata=v0_item.nodata,
            attrs=v0_item.attrs,
        )

        adapter_dict = {}

        if v0_item.rename:
            adapter_dict["rename"] = v0_item.rename

        if v0_item.unit_add:
            adapter_dict["unit_add"] = v0_item.unit_add

        if v0_item.unit_mult:
            adapter_dict["unit_mult"] = v0_item.unit_mult

        if len(adapter_dict) > 0:
            data_adapter = DataCatalogV1DataAdapter(**adapter_dict)
        else:
            data_adapter = None

        return DataCatalogV1Item(
            name=v0_item.name,
            data_type=v0_item.data_type,
            uri=uri,
            data_adapter=data_adapter,
            driver=driver,
            metadata=metadata,
        )

    @model_validator(mode="after")
    def uri_in_item_or_variants(self):
        if self.uri is not None or (
            self.variants is not None
            and all(variant.uri is not None for variant in self.variants)
        ):
            return self
        else:
            raise ValueError(
                "Source must either have a uri, or all of it's variants must have one."
            )

    @staticmethod
    def from_dict(input_dict, name=None):
        """Convert a dictionary into a validated source item."""
        dict_name = input_dict.pop("name", None)
        if name is None:
            entry_name = dict_name
        else:
            entry_name = name
        try:
            item_metadata = DataCatalogV1ItemMetadata.from_dict(
                input_dict.pop("metadata", {})
            )

            return DataCatalogV1Item(
                **input_dict,
                name=entry_name,
                metadata=item_metadata,
            )
        except ValidationError as e:
            raise ValidationError.from_exception_data(
                entry_name or "nameless entry", e.errors(), "python"
            )


class DataCatalogV1Validator(BaseModel):
    """A validated complete data catalog."""

    meta: DataCatalogV1MetaData | None = None
    sources: Dict[str, DataCatalogV1Item] = Field(default_factory=dict)

    model_config = ConfigDict(
        str_strip_whitespace=True,
        extra="forbid",
    )

    @model_serializer()
    def serialize_model(self):
        if self.meta:
            out = {
                "meta": self.meta.model_dump(),
            }
        else:
            out = {}

        serialized_sources = {
            key: source.model_dump(
                exclude_unset=True, exclude_none=True, exclude={"name"}
            )
            for (key, source) in self.sources.items()
        }
        return {**out, **serialized_sources}

    @staticmethod
    def from_v0(v0_catalog: DataCatalogV0Validator) -> "DataCatalogV1Validator":
        if v0_catalog.meta:
            v1_meta = DataCatalogV1MetaData.from_v0(v0_catalog.meta)
        else:
            v1_meta = None

        converted_entries = {}

        for entry_id, entry in v0_catalog.sources.items():
            converted_entry = DataCatalogV1Item.from_v0(entry)
            converted_entries[entry_id] = converted_entry

        return DataCatalogV1Validator(meta=v1_meta, sources=converted_entries)

    @staticmethod
    def from_dict(input_dict):
        """Create a validated datacatalog from a dictionary."""
        if input_dict is None:
            return DataCatalogV1Validator()
        else:
            if meta := input_dict.pop("meta", None):
                catalog_meta = DataCatalogV1MetaData.from_dict(meta)
            else:
                catalog_meta = None

            catalog_entries = {}
            for entry_name, entry_dict in input_dict.items():
                catalog_entries[entry_name] = DataCatalogV1Item.from_dict(
                    entry_dict, name=entry_name
                )

            return DataCatalogV1Validator(meta=catalog_meta, sources=catalog_entries)

    @staticmethod
    def from_yml(path: str):
        """Create a validated data catalog loaded from the provided path."""
        yml_dict = _yml_from_uri_or_path(path)
        return DataCatalogV1Validator.from_dict(yml_dict)



"""Abstract driver to read datasets."""

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import xarray as xr

from hydromt.data_catalog.drivers.base_driver import (
    BaseDriver,
)
from hydromt.error import NoDataStrategy

logger = logging.getLogger(__name__)


class DatasetDriver(BaseDriver, ABC):
    """Abstract Driver to read Datasets."""

    @abstractmethod
    def read(
        self, uris: list[str], *, handle_nodata: NoDataStrategy = NoDataStrategy.RAISE
    ) -> xr.Dataset:
        """
        Read data from one or more URIs into an xarray Dataset.

        This abstract method defines the interface for all dataset drivers. Subclasses
        should implement data loading logic appropriate for the format being read.

        Parameters
        ----------
        uris : list[str]
            List of URIs to read data from.
        handle_nodata : NoDataStrategy, optional
            Strategy to handle missing or empty data. Default is NoDataStrategy.RAISE.

        Returns
        -------
        xr.Dataset
            The loaded dataset.
        """
        ...

    @abstractmethod
    def write(
        self,
        path: Path | str,
        data: xr.Dataset,
        *,
        write_kwargs: dict[str, Any] | None = None,
    ) -> Path:
        """
        Write an xarray Dataset to disk.

        This abstract method defines the interface for all Dataset-based drivers.
        Subclasses should implement logic for writing datasets in specific formats
        (e.g., NetCDF, Zarr).

        Parameters
        ----------
        path : Path | str
            Destination path or URI where the Dataset should be written.
        data : xr.Dataset
            The Dataset to write.
        write_kwargs : dict[str, Any], optional
            Additional keyword arguments to pass to the underlying xarray write function
            (e.g., `to_zarr`, `to_netcdf`). Default is None.

        Returns
        -------
        Path
            The path where the data was written.

        """
        ...



"""The GeoDataFrame adapter performs transformations on GeoDataFrames."""

import logging
from typing import Any, Dict, Iterable, List, Optional, Union

import geopandas as gpd
import numpy as np
import pyproj
from pyproj import CRS

from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.gis.vector_utils import _filter_gdf
from hydromt.typing import (
    Geom,
    SourceMetadata,
)

logger = logging.getLogger(__name__)


class GeoDataFrameAdapter(DataAdapterBase):
    """The GeoDataFrameAdapter performs transformations on GeoDataFrames."""

    def transform(
        self,
        gdf: gpd.GeoDataFrame,
        metadata: "SourceMetadata",
        *,
        mask: Optional[gpd.GeoDataFrame] = None,
        variables: Optional[List[str]] = None,
        predicate: str = "intersects",
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[gpd.GeoDataFrame]:
        """Read transform data to HydroMT standards.

        Parameters
        ----------
        gdf : gpd.GeoDataFrame
            input GeoDataFrame
        metadata : SourceMetadata
            source metadata
        mask : Optional[gpd.GeoDataFrame], optional
            mask to filter by geometry, by default None
        variables : Optional[List[str]], optional
            variable filter, by default None
        predicate : str, optional
            predicate to use for the mask filter, by default "intersects"
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE

        Returns
        -------
        Optional[gpd.GeoDataFrame]
            filtered and harmonized GeoDataFrame

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        # rename variables and parse crs & nodata
        gdf = self._rename_vars(gdf)
        gdf = self._set_crs(gdf, metadata.crs)
        gdf = self._set_nodata(gdf, metadata)
        # slice
        gdf: Optional[gpd.GeoDataFrame] = GeoDataFrameAdapter._slice_data(
            gdf,
            variables=variables,
            mask=mask,
            predicate=predicate,
            handle_nodata=handle_nodata,
        )
        if gdf is None:
            return None

        # uniformize
        gdf = self._apply_unit_conversions(gdf)
        gdf = self._set_metadata(gdf, metadata)
        return gdf

    def _rename_vars(self, gdf: gpd.GeoDataFrame):
        # rename and select columns
        if rename := self.rename:
            rename = {k: v for k, v in rename.items() if k in gdf.columns}
            gdf = gdf.rename(columns=rename)
        return gdf

    def _set_crs(self, gdf: gpd.GeoDataFrame, crs: Optional[CRS]):
        if crs is not None and gdf.crs is None:
            gdf.set_crs(crs, inplace=True)
        elif gdf.crs is None:
            raise ValueError("GeoDataFrame: CRS not defined in data catalog or data.")
        elif crs is not None and gdf.crs != pyproj.CRS.from_user_input(crs):
            logger.warning(
                "GeoDataFrame : CRS from data catalog does not match CRS of"
                " data. The original CRS will be used. Please check your data catalog."
            )
        return gdf

    @staticmethod
    def _slice_data(
        gdf: gpd.GeoDataFrame,
        variables: Optional[Union[str, List[str]]] = None,
        mask: Optional[Geom] = None,
        predicate: str = "intersects",
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[gpd.GeoDataFrame]:
        """Filter the GeoDataFrame.

        Parameters
        ----------
        gdf : gpd.GeoDataFrame
            _description_
        variables : Optional[Union[str, List[str]]], optional
            variables to include, all when None, by default None
        mask : Optional[Geom], optional
            filter by geometry, or keep all if None, by default None
        predicate : str, optional
            predicate to use for the geometry filter, by default "intersects"
            should be one of the Shapeley binary predicates
            (https://shapely.readthedocs.io/en/latest/manual.html#binary-predicates)
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE

        Returns
        -------
        Optional[gpd.GeoDataFrame]
            filtered GeoDataFrame, or None if no data remains

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        if variables is not None:
            variables = np.atleast_1d(variables).tolist()
            if np.any([var not in gdf.columns for var in variables]):
                raise ValueError(f"GeoDataFrame: Not all variables found: {variables}")
            if "geometry" not in variables:  # always keep geometry column
                variables = variables + ["geometry"]
            gdf = gdf.loc[:, variables]
            if np.all(gdf.is_empty):
                exec_nodata_strat(
                    "GeoDataFrame has no data after variable slicing.", handle_nodata
                )
                return None

        if mask is not None:
            # NOTE if we read with vector driver this is already done ..
            bbox_str = ", ".join([f"{c:.3f}" for c in mask.total_bounds])
            epsg = mask.crs.to_epsg()
            logger.debug(f"Clip {predicate} [{bbox_str}] (EPSG:{epsg})")
            idxs = _filter_gdf(gdf, geom=mask, predicate=predicate)
            gdf = gdf.iloc[idxs]
            if np.all(gdf.is_empty):
                exec_nodata_strat(
                    "GeoDataFrame has no data after masking.", handle_nodata
                )
                return None

        return gdf

    def _set_nodata(self, gdf: gpd.GeoDataFrame, metadata: "SourceMetadata"):
        """Parse and apply nodata values from the data catalog."""
        cols: Iterable[str] = gdf.select_dtypes([np.number]).columns
        no_data_value: Union[Dict[str, Any], str, None]
        if no_data_value := metadata.nodata is not None and len(cols) > 0:
            if not isinstance(no_data_value, dict):
                no_data_dict: Dict[str, Any] = {c: no_data_value for c in cols}
            else:
                no_data_dict: Dict[str, Any] = no_data_value
            for c in cols:
                mv = no_data_dict.get(c, None)
                if mv is not None:
                    is_nodata = np.isin(gdf.loc[:, c], np.atleast_1d(mv))
                    gdf.loc[:, c] = np.where(is_nodata, np.nan, gdf.loc[:, c])
        return gdf

    def _apply_unit_conversions(self, gdf: gpd.GeoDataFrame):
        # unit conversion
        unit_names = list(self.unit_mult.keys()) + list(self.unit_add.keys())
        unit_names = [k for k in unit_names if k in gdf.columns]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} columns.")
        for name in list(set(unit_names)):  # unique
            m = self.unit_mult.get(name, 1)
            a = self.unit_add.get(name, 0)
            result = gdf.loc[:, name] * m + a
            if not np.issubdtype(gdf[name].dtype, np.floating):
                result = result.astype(gdf[name].dtype)  # cast to original dtype
            gdf.loc[:, name] = result
        return gdf

    def _set_metadata(self, gdf: gpd.GeoDataFrame, metadata: "SourceMetadata"):
        # .`attrs` is not a stable part of the (geo)pandas API, see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.attrs.html#pandas.DataFrame.attrs
        # There are some issues and bug reports from its usage that are very similar to this.
        # (Geo)Pandas' current recommendation is to avoid setting attrs on series within a (Geo)DataFrame and only have one global attrs dict per (Geo)DataFrame.
        # https://github.com/pandas-dev/pandas/issues/35425
        # https://github.com/dask/dask/pull/6742
        # https://github.com/pandas-dev/pandas/issues/41572
        # https://stackoverflow.com/questions/72803056/why-do-attrs-of-series-in-a-pandas-dataframes-change-to-the-dataframe-attrs
        meta = metadata.model_dump()
        attrs = meta.pop("attrs", {})
        for k, v in attrs.items():
            meta[k] = v
        gdf.attrs.update(**meta)

        return gdf



"""Decorator for hydromt steps."""

from typing import Any, Callable, TypeVar

F = TypeVar("F", bound=Callable[..., Any])


def hydromt_step(funcobj: F) -> F:
    """Decorate a method indicating it is a hydromt step.

    Only methods decorated with this decorator are allowed to be called by Model.build and Model.update.
    """
    funcobj.__ishydromtstep__ = True  # type: ignore[attr-defined]
    return funcobj
